luanet.load_assembly("System");
luanet.load_assembly("System.Xml");
luanet.load_assembly("log4net");

local Types = {};
Types["System.Net.WebClient"] = luanet.import_type("System.Net.WebClient");
Types["System.Text.Encoding"] = luanet.import_type("System.Text.Encoding");
Types["log4net.LogManager"] = luanet.import_type("log4net.LogManager");

require "JsonWriter";
require "JsonParser";

local Settings = {};
Settings.InternetArchiveApiBaseUrl = GetSetting("InternetArchiveApiBaseUrl");
Settings.InternetArchiveWebPlatformConfigId = GetSetting("InternetArchiveWebPlatformConfigId");
Settings.MonitoredQueue = GetSetting("MonitoredQueue");
Settings.ReviewQueue = GetSetting("ReviewQueue");
Settings.ErrorQueue = GetSetting("ErrorQueue");

local rootLogger = "AtlasSystems.Addons.InternetArchive";
local log = Types["log4net.LogManager"].GetLogger(rootLogger);

local isCurrentlyProcessing = false;

function Init()
    RegisterSystemEventHandler("SystemTimerElapsed", "ProcessInternetArchiveRequests");
end

function ProcessInternetArchiveRequests()
	if not isCurrentlyProcessing then
		isCurrentlyProcessing = true;

		local success, error = pcall(function() ProcessDataContexts("TransactionStatus", Settings.MonitoredQueue, "SubmitInternetArchiveRequest") end);
		if success then
			log:Info("Successfully processed requests for Internet Archive.");
		else
			log:Error("An error occurred while processing requests for Internet Archive.");
			OnError(error);
		end

		isCurrentlyProcessing = false;
	else
		log:Info("Requests are still being processed for Internet Archive.");
	end
end

function SubmitInternetArchiveRequest()
	local transactionNumber = GetFieldValue("Transaction", HostAppInfo.SourceFields["TransactionNumber"]);

	local requestTable = BuildRequestTable();
	if not requestTable then
		LogNoteAndRoute(transactionNumber, "Problem building request table for Internet Archive", Settings.ErrorQueue);
	end

	local requestJson = JsonWriter:Serialize(requestTable);

	local response = PostToInternetArchiveApi(requestJson);
	if not response then
		LogNoteAndRoute(transactionNumber, "Failed to send Internet Archive request", Settings.ErrorQueue);
		return;
	end

	local responseTable = JsonParser:ParseJSON(response);
	if responseTable["response"] == "maybe" then
		LogNoteAndRoute(transactionNumber, "Internet Archive request successfully placed", "Request Sent");
	elseif responseTable["response"] == "no" then
		LogNoteAndRoute(transactionNumber, "Request cannot be supplied by Internet Archive", Settings.ReviewQueue);
	else
		LogNoteAndRoute(transactionNumber, "Failed to send Internet Archive request", Settings.ErrorQueue);
	end
end

function BuildRequestTable()
	log:Debug("Building request table");

	local requestTable = {};

	requestTable["Client"] = "Atlas";

	for k,v in pairs(HostAppInfo.SourceFields) do
		requestTable[k] = GetFieldValue("Transaction", v);
	end

	local apiInfo = QueryDatabase("SELECT BaseWebServiceUrl, CONVERT(nvarchar(50), ApiKey) AS ApiKey FROM WebPlatformConfig WHERE ID = '" .. Settings.InternetArchiveWebPlatformConfigId .. "'");
	if not apiInfo then
		log:Error("Unable to retrieve info from WebPlatformConfig");
		return false;
	end

	requestTable["ILLiadAPIURL"] = apiInfo.Rows:get_Item(0):get_Item("BaseWebServiceUrl");
	requestTable["ILLiadAPIKey"] = apiInfo.Rows:get_Item(0):get_Item("ApiKey");

	local nvtgc = ILLiad.GetParentNVTGC(GetFieldValue("User", "NVTGC"));
	requestTable["OdysseyAddress"] = QueryDatabase("SELECT Value FROM Customization WHERE CustKey = 'OdysseySystemId' AND NVTGC = '" .. nvtgc .. "'", true);
	if not requestTable["OdysseyAddress"] then
		log:Error("Unable to retrieve OdysseySystemId for " .. nvtgc);
		return false;
	end

	requestTable["LibraryID"] = QueryDatabase("SELECT Value FROM Customization WHERE CustKey = 'SystemIDOCLCSymbol' AND NVTGC = '" .. nvtgc .. "'", true);
	if not requestTable["LibraryID"] then
		log:Error("Unable to retrieve SystemIDOCLCSymbol for " .. nvtgc);
		return false;
	end

	return requestTable;
end

function PostToInternetArchiveApi(json)
	log:Debug("Sending request to Internet Archive API");

	local client = Types["System.Net.WebClient"]();
    client.Encoding = Types["System.Text.Encoding"].UTF8;
	local url = Settings.InternetArchiveApiBaseUrl;

	local function UploadString(url, json)
		client.Headers:Clear();
		client.Headers:Add("accept", "application/json");
		client.Headers:Add("Content-Type", "application/json");

		local results = client:UploadString(url, json);
		return results;
	end
	
	local success, results = pcall(UploadString, url, json);

	client:Dispose();

	if not success then
		OnError(results);
		return false;
	end
	
	return results;
end

function LogNoteAndRoute(transactionNumber, note, queue)
	log:Debug("Transaction " .. tostring(transactionNumber) .. " - " .. tostring(note) .. " - Routing to " .. tostring(queue));
	ExecuteCommand("AddNote", {transactionNumber, "Internet Archive Addon - " .. note});
	ExecuteCommand("Route", {transactionNumber, queue});
end

function QueryDatabase(query, isScalar)
	local connection = CreateManagedDatabaseConnection();
	local function DbCall()
		connection.QueryString = query;
		connection:Connect();

		local results;
		if isScalar then
			results = connection:ExecuteScalar();
		else
			results = connection:Execute();
		end
		
		return results;
	end

	local success, results = pcall(DbCall, query);

	connection:Dispose();

	if not success then
		return false;
	end
	
	return results;
end

function OnError(e)
    if e == nil then
        log:Error("OnError supplied a nil error");
        return;
    end

    if not e.GetType then
        -- Not a .NET type. Attempt to log value
        pcall(function ()
            log:Error(e);
        end);
        return;
    else
        if not e.Message then
            log:Error(e:ToString());
            return;
        end
    end

    local message = TraverseError(e);

    if message == nil then
        message = "Unspecified Error";
    end

    log:Error("An error occurred while processing the Internet Archive request:\r\n" .. message);
end

function TraverseError(e)
	-- Recursively logs exception messages and returns the innermost message to caller	
    if not e.GetType then
        -- Not a .NET type
        return nil;
    else
        if not e.Message then
            -- Not a .NET exception
            log:Error(e:ToString());
            return nil;
        end
    end

    log:Error(e.Message);

    if e.InnerException then
        return TraverseError(e.InnerException);
    else
        return e.Message;
    end
end

function GetErrorMessage(e)
	log:Debug("Getting error message");
    if not e.Message then
		-- Not a .NET exception
		log:Debug("Not a .NET exception");
		return tostring(e);
	end

    if e.InnerException then
        return e.Message .. "\r\n" .. GetErrorMessage(e.InnerException);
    else
        return e.Message;
    end
end