--[[
JsonWriter.lua

Summary:

    This module accepts lua tables and returns matching JSON objects or arrays, 
    formatted as strings. The module supports nested objects and nested arrays. 
    The output will be minified. 

Usage:

    ```lua
    local example = {
        key1 = "test",
        key2 = {
            keyA = "val1"
        },
        key3 = {
            "val1",
            "val2"
        }
    }

    local json = JsonWriter:Serialize(example)
    LogDebug(json)

    -- Result:
    -- {"key1":"test","key2":{"keyA":"val1"},"key3":["val1","val2"]}
    ```

Remarks:

    There is no way to differentiate empty "objects" and empty "arrays" in lua.
    Any empty tables that exist in the input data will be serialized as empty
    arrays.

.NET Library Requirements:

    None

]]

JsonWriter = {}
JsonWriter.__index = JsonWriter
JsonWriter.NIL = {}

JsonWriter.Types = {}

function JsonWriter:Serialize (value)
    --[[

        Serializes a value as a string, containing a JSON representation of the
        value, regardless of the value's type.

    ]]

    if type(value) == "table" then
        if (next(value) == nil) or value[1] then
            return JsonWriter:SerializeArray(value)
        end

        return JsonWriter:SerializeTable(value)
    end

    return JsonWriter:FormatValue(value)
end

function JsonWriter:SerializeTable (luaTable)
    --[[

        Serializes a lua table as a string containing a JSON representation of 
        the table. 

    ]]

    local objectParameters = {}

    for k,v in pairs(luaTable) do
        local val = JsonWriter:Serialize(v)
        table.insert(objectParameters, string.format("\"%s\":%s", k, val))
    end

    return string.format("{%s}", table.concat(objectParameters, ","))
end

function JsonWriter:SerializeArray (luaArray)
    --[[

        Serializes a lua array as a string containing a JSON representation of 
        the array. 

    ]]

    local arrayValues = {}

    for i,v in ipairs(luaArray) do
        local val = JsonWriter:Serialize(v)
        table.insert(arrayValues, val)
    end

    return string.format("[%s]", table.concat(arrayValues, ","))
end

function JsonWriter:FormatValue (value)
    return (
        (type(value) == "nil" and "null") or
        (type(value) == "boolean" and tostring(value)) or
        (type(value) == "number" and tostring(value)) or
        string.format('"%s"', value:gsub('["%\\%z\001-\031]', function (c)
            return (
                (c == "\n" and "\\n") or
                (c == "\r" and "\\r") or
                (c == "\t" and "\\t") or
                (c == "\b" and "\\b") or
                (c == "\f" and "\\f") or
                (c == '"' and '\\"') or
                (c == '\\' and '\\\\') or
                string.format("\\u%04x", c:byte())
            )
        end))
    )
end
